/*
 * cancan  --  mud client with telnet protocol
 *
 * written by Mattias Engdegrd (Yorick) (f91-men@nada.kth.se) 1992-95
 * Maintainers: Ilie (finnag@pvv.unit.no) (and Yorick)
 * originally inspired by the Tintin client by Peter Unold but contains no
 * Tintin code.
 * Also greatly improved upon by Vivriel and Thuzzle. As usual, all the
 * developers are in debt to countless users for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define VERSION "2.6.5 (Yorick, Vivriel, Thuzzle, Ilie 1992-97)"
#define SAVEFILEVER "2.6"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"
#include "beam.h"

int select();
int gettimeofday();

/*	function declarations		*/
static void init_def_file();
static void mainlupe();
static void process_attn();
#ifndef OMIT_SCHEDULE
static struct timeval *check_schedule();
#endif
#ifdef MOTDFILE
static void printmotd();
#endif
static void get_remote_output();
#ifndef OMIT_SESSIONS
static void process_sub_input();
#endif
static void process_input();
static void getuserinput();
static void setsigs();
static void init_env();
static void read_settings();
void termsig();
void pipesig();
void childsig();
void winchsig();

/*
 * A significant overhead could occur in strncmp when variables are
 * heavily used. Rather, we use a hash value to speed things up.
 */
static char hash_var();
static int find_env();

static char *envhash;
static int envsize;
static int envtop;

/*	GLOBALS		*/

int portnumber;
char *hostname;
char *initstr;		/* initial command to run (send) on connect */

FILE *capfile;		/* capture file or NULL */

#ifndef OMIT_SESSIONS
/* session data indexed by socket descriptor */
static char *sess_id[MAXFDSCAN];	/* session name */
static int sess_active[MAXFDSCAN];	/* active flag */
static int main_sockfd;	/* main session socket fd */
#endif

int sockfd;		/* current socket file descriptor */
fd_set fdset;		/* set of descriptors to select() on */

int linemode;		/* line mode flags (LM_* in cancan.h) */
int attention;		/* special attention flags (ATTN_* in cancan.h) */

/* booleans */
char current_options[] = OPTION_DEFAULTS;
int mp_active;		/* 1 = we have #identified ourselves */
int expect_prompt;	/* are we waiting for remote data? */
int new_settings;	/* set this when user defines new alias/action/etc */
int verbose;		/* 1 if command being parsed is user-entered */

int edlen;		/* length of current input line */
int pos;		/* cursor position in line */
int flashback;		/* cursor is on excursion, put back there */
char *prefixstr;	/* inserted in the editing buffer each time */
char *footnote;		/* displayed at the bottom (just once) */
char edbuf[LINELEN];	/* line editing buffer */
char promptstr[LINELEN];/* entire prompt string (for redrawing) */
char insertbuf[LINELEN];/* used in many ways. hm */

int main(argc, argv)
int argc; char *argv[];
{
    static char outbuf[BUFSIZE];
    if(argc < 2 || argc > 3) {
	fprintf(stderr, "usage:\t%s host port\nor\t%s definition-file\n",
		argv[0], argv[0]);
	return 1;
    }
    /* full buffered stdout means less write() calls */
    setvbuf(stdout, outbuf, _IOFBF, sizeof(outbuf));
    init_env();		/* set up our environment first */
#if !defined(VT100) || defined(TELNETBUG)
    settermcap();
#endif
    default_bindings();
#ifndef OMIT_PRINTVER
    printver();
#endif
#ifdef MOTDFILE
    printmotd();
#endif
    if(argc == 2) {	/* one arg, assuming it is definition file name */
	init_def_file(argv[1]);
    } else {	/* two args: assume they are hostname and port number */
	hostname = argv[1];	/* should be safe not to str_dup it */
	portnumber = atoi(argv[2]);
	keypad_bindings();
    }
    default_completions();
    set_terminal();
    setsigs();
    mainlupe();
    return 1;
}

/*
 * initialize cancan from definition file or create one
 */
static void init_def_file(df)
char *df;
{
    char buf[LINELEN], *p;
    FILE *f;
    /*
     * rules: if cancan dir is set it is searched first.
     *        if file doesn't exist, it is created there.
     * If a slash appears in the name, the cancan dir isn't used.
     */
    if(!strchr(df, '/') && (p = get_env("CANCANDIR")) && *p) {
	/* BUG: buffer overflow possible. <snicker> */
	sprintf(buf, "%s/%s", p, df);
	if((f = fopen(buf, "r")) || !(f = fopen(df, "r")))
	    df = buf;
    } else
	f = fopen(df, "r");
    set_env("SAVEFILE", df);
    if(f) {
	read_settings(f);
	if(ferror(f)) 
	    exit(syswarn("Can't read definition file"));
	fclose(f);
    } else {
	if(!(f = fopen(df, "a+")))
	    exit(syswarn("Cannot open definition file"));
	/* unlink(df); */
	fclose(f);
	printf("Creating %s\nHost name  :", df);
	fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	if(*skipspace(buf) != '\n') {
	    chop_str(buf, '\n');  hostname = str_dup(buf);
	    printf("Port number:");
	    fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	    portnumber = atoi(buf);
	}
	keypad_bindings();
	save_settings();
    }
}

/*
 * show current version
 */
void printver()
{
    printf(
#ifdef __STDC__
	"cancan ver " VERSION "\nCompiled " __DATE__ " (%s %s)\n",
#else
	"cancan ver %s (%s %s)\n", VERSION,
#endif
#ifdef VT100
	"vt100-only",
#else
	"termcaps",
#endif
#ifndef USESGTTY
	"termio"
#else
	"BSD sgtty"
#endif
    );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
static void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if(f) {
	while(fgets(line, sizeof(line), f))
	    fputs(line, stdout);
	fclose(f);
    }
}
#endif

/*
 * main loop.
 */
static void mainlupe()
{
    int err;
    fd_set readfds;
    struct timeval *timeout;
    static struct timeval timeoutbuf;

    FD_ZERO(&fdset);
    FD_SET(0, &fdset);		/* standard input (terminal) */
#ifndef OMIT_SESSIONS
    if(hostname && connect_sess((char *)0, initstr, (char *)0, 0) < 0)
	exitcancan("");
#else
    if(!hostname || (sockfd = doconnect(hostname, portnumber)) < 0)
	exitcancan("#please specify correct hostname and port number\n");
    FD_SET(sockfd, &fdset);	/* add the socket to select() set */
    if(initstr) parse(initstr, 0, (struct param *)0);
#endif

    printf("Type #help for help.\n");

    /* make sure the input line is always the bottom one */
    check_winsize();
    lastline();

    /*		main lupe	*/

    for(;;) {

#ifndef OMIT_SESSIONS
	sockfd = main_sockfd;
#endif

#ifndef OMIT_SCHEDULE
	timeout = check_schedule(&timeoutbuf);
#else
	timeout = 0;
#endif

	if(attention & (ATTN_FLASHBACK | ATTN_FOOTNOTE)) {
	    if(attention & ATTN_FLASHBACK)
#ifndef OMIT_SCHEDULE
		if(!timeout || timeout->tv_sec)
#endif
		{
		    timeout = &timeoutbuf,
		    timeout->tv_sec = FLASHDELAY / 1000000,
		    timeout->tv_usec = FLASHDELAY % 1000000;
		}
	    if(attention & ATTN_FOOTNOTE)
		display_footnote();
	}

	flush_output_buffer();	/* sploooosh */
	fflush(stdout);		/* ditto */

	/*
	 * NOTE: we should pass highest_fd + 1, rather than MAXFDSCAN
	 * NOTE: check_schedule gettimeofday should come right before select
	 * NOTE: a signal arriving just before we enter select has to await
	 * some other event, before it gets processed... is there any way
	 * around this without chemistry?
	 */
	readfds = fdset;
	err = select(MAXFDSCAN, &readfds, (fd_set *)0, (fd_set *)0, timeout);
	if(err < 0 && errno != EINTR) syserr("select");

	if(attention) process_attn();

	if(err > 0) {
#ifndef OMIT_SESSIONS
	    if(FD_ISSET(0, &readfds)) {
		getuserinput();
		if(!--err) continue;
	    }
	    do if(FD_ISSET(sockfd, &readfds)) {
		get_remote_output();
		if(!--err) break;
	    } while(++sockfd < MAXFDSCAN);
#else
	    if(FD_ISSET(0, &readfds))
		getuserinput();
	    if(FD_ISSET(sockfd, &readfds))
		get_remote_output();
#endif
	}
    }
}

static void process_attn()
{
    int attn = attention;
    attention ^= attn;
    if(attn & ATTN_FLASHBACK) {
	moveto(flashback);
	if(!(attn & ~ATTN_FLASHBACK)) return;
    }
    if(attn & ATTN_FOOTNOTE) {
	str_assign(&footnote, (char *)0);
	display_footnote();
	if(!(attn & ~(ATTN_FLASHBACK | ATTN_FOOTNOTE))) return;
    }
    if(attn & ATTN_WINCHSIG) {
	signal(SIGWINCH, winchsig);
	if(check_winsize()) {
	    clear_all();
	    redraw_all();
	}
    }
    if(attn & ATTN_CHILDSIG) {
	/*
	 * NOTE: on some systems (e.g. AIX), setting a sigchld handler
	 * will promptly deliver the signal if waitpid() has anything to
	 * report. Thus, we would get called once more on the next loop,
	 * and do extra signal() and waitpid() calls. Better flush one
	 * child before reinstalling the signal handler.
	 */
	attend_child(0);
	signal(SIGCHLD, childsig);
	attend_child(-1);
    }
}

#ifndef OMIT_SCHEDULE
/*
 * process and update delayed commands list.
 */
static struct timeval *check_schedule(to)
struct timeval *to;
{
    static struct timeval time0;
    if(schedule) {
	long delta = -time0.tv_sec;
	/*
	 * NOTE: gettimeofday is somewhat expensive. However, even if we
	 * used alarm timer, the chedule still needs to be up to date,
	 * in case user decides to list it.
	 */
	gettimeofday(&time0, (struct timezone *)0);
	time0.tv_sec = SCH_HZ * time0.tv_sec +
		time0.tv_usec / (1000000 / SCH_HZ);
	if(delta) delta += time0.tv_sec;
	/*
	 * NOTE: when a command is run while linemode is LM_CHAR,
	 * the output produced (if any) might not look too good since
	 * LF to CRLF mapping is not in effect. This is rare, though.
	 * Besides, in char mode user cannot change the schedule;
	 * perhaps check_schedule should be disabled then.
	 */
	do {
	    if(schedule->delta > delta) {
		delta = (schedule->delta -= delta);
		to->tv_sec = delta / SCH_HZ;
		to->tv_usec = (1000000 / SCH_HZ) * (delta % SCH_HZ);
		return to;
	    } else if(schedule->delta >= 0) {
		delta -= schedule->delta;
		schedule->delta = -1;
		if(schedule->next)
		    schedule->next->delta++;
		run_command(schedule->commands);
	    } else
		delete_schedulenode(&schedule);
	} while(schedule);
    }
    time0.tv_sec = 0;
    return 0;
}
#endif

/*
 * read terminal input and send to parser.
 */
static void getuserinput()
{
    char rbuf[CAPLEN];
    int n = read(0, rbuf, sizeof(rbuf));
    if(n <= 0)
	syserr("read from stdin");
    if(!(linemode & LM_CHAR)) {
	line_edit(rbuf, n);
	return;
    }
    /* character mode, no line editing */
    while(write(sockfd, rbuf, n) < 0)
	if(errno != EINTR) { syserr("write to socket"); break; }
    if(!(linemode & LM_NOECHO))
	while(write(1, rbuf, n) < 0)
	    if(errno != EINTR) { syserr("write to stdout"); break; }
}

/*
 * get output from the socket and process/display it.
 */
static void get_remote_output()
{
    static char buffer[BUFSIZE];
    int got = read_socket(buffer, sizeof(buffer) - 1);
    /* leave room for terminating zero:          ^^^ */
    if(!got) return;
#ifdef DEBUGCODE
    /* debug code to see in detail what strange codes come from the server */
    {
	int i;
	for(i = 0; i < got; i++)
	    fprintf(stderr, isprint(buffer[i]) ? "%c" : "[%d]", buffer[i]);
    }
#endif

    if(got <  0) {
#ifndef OMIT_SESSIONS
	/* NOTE: What if we are in LM_CHAR mode??  */
	clear_all();
	disconnect_sess(sockfd);
	redraw_all();
	return;
#else
	exitcancan("#connection closed by remote host.\n");
#endif
    }
#ifndef OMIT_SESSIONS
    if(!sess_active[sockfd]) return;  /* process only active sessions */
#endif
    if(!(linemode & LM_CHAR)) {
	/* line-at-a-time mode: process input in a number of ways */
	clear_input();
#ifndef OMIT_SESSIONS
	if(sockfd != main_sockfd) process_sub_input(buffer, got);
	else
#endif
	process_input(buffer, got);
	redraw_input();
    } else {
	while(write(1, buffer, got) < 0)
	    if(errno != EINTR) { syserr("write to stdout"); break; }
    }
}

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to stdout
 * assert(sizeof(buf) >= size + 1)
 */
static void process_input(buf, size)
char *buf; int size;
{
    static editsess *esess;
    char *p;
    int l, i = 0;

    if(expect_prompt) {
	/* NOTE: expect_prompt is zeroed even if we get only a mp message */
	expect_prompt = 0;
	if(esess) esess->mode = 0;	/* sending cancels the session */
    }
    while(i < size) {
	p = buf + i;
	l = size - i;
	if(esess || (p[0] == MPI[0] && mp_active && !col0 && IS_MPI(p, l)))
	    i += mp_receive(&esess, p, l);
	else {
	    /* process block one line at a time: */
	    if(*p == '\r' && !col0) i++;		/* leading CR */
	    for(p = buf; i < size && buf[i] != '\n'; i++)
		if(buf[i]) *p++ = buf[i];		/* cut out NUL's */
	    if(p != buf && *(p - 1) == '\r') p--;	/* trailing CR */
	    *p = 0;
	    if(capfile) to_capfile(buf, (i < size));
	    if(i++ < size) {
		if(!search_action(buf, p - buf))	/* line #gagged */
		    continue;
		if(buf[0]) wrap_out(buf);
		newline();  col0 = 0;
	    } else {
		/*
		 * FEATURE: action search on line fragments can give results
		 * differing from a search on a complete line. Alternatively,
		 * the search could be postponed until a newline is read.
		 */
		int old_prompt;
		search_action(buf, p - buf);
		old_prompt = col0;
		p = wrap_out(buf);
		if(p) old_prompt = 0;
		else p = buf;
		if(!old_prompt) promptstr[0] = 0;
		if(col0 < cols_1 &&
			strlen(p) + (l = strlen(promptstr)) < LINELEN) {
		    strcpy(promptstr + l, p);		/* promptocatenate */
		} else {
		    newline();  col0 = 0;
		}
		break;
	    }
	}
    } /* end while */
}

#ifndef OMIT_SESSIONS
/*
 * process input from remote host (subsidiary session):
 * assert(sizeof(buf) >= size + 1)
 */
static void process_sub_input(buf, size)
char *buf; int size;
{
    char *p;
    int i;
    int back_col0;
    clear_prompt();
    back_col0 = col0; col0 = 0;
    /*
     * FEATURE: we don't keep track of subsession prompt, nor subsession
     * col0 (could use an array for that). So we set col0 to zero when
     * searching actions to trigger. Therefore, '^' anchoring an action
     * makes no difference for subsessions. (Subsession handling really
     * needs a rewrite. At least there's a lot of room for improvement.)
     */
    for(i = 0; i < size; i++) {
	if(buf[i] == '\r') continue;
	for(p = buf; i < size && buf[i] != '\n'; i++)
	    if(buf[i]) *p++ = buf[i];	/* cut out NUL's */
	*p = 0;		/* there is room */
	/*
	 * FEATURE: no wrapping, no marking in subsidiary session
	 * nonprintable chars (escape codes) should be removed too
	 */
	if(search_action(buf, p - buf)) {
#if defined(_POSIX_SOURCE)
	    printf("%.*s", cols_1 - printf("##%s> ", sess_id[sockfd]), buf);
#else
	    printf("##%s> %.*s", sess_id[sockfd],
		    cols_1 - (int)(4 + strlen(sess_id[sockfd])), buf); 
#endif
	    newline();
	}
    }
    col0 = back_col0;
    redraw_prompt();
}
#endif

#ifndef NOSHELL
/*
 * execute a shell command and return the result
 */
int execute(arg)
char *arg;
{
    int pid;
    flush_output_buffer();
    fflush(stdout);
    pid = fork();
    if(!pid) {
	static char *args[] = { "/bin/sh", "-c", 0, 0 };
	args[2] = arg;
	execve(args[0], args, environ);
	_exit(1);
    } else if(pid < 0) {
	printf("#fork failed, sorry\n");
    } else {
	pid = attend_child(pid);
    }
    return pid;		/* return the process exit status */
}
#endif

/*
 * set up our signal handlers
 */
static void setsigs()
{
    signal(SIGTERM, termsig);
    signal(SIGPIPE, pipesig);
    signal(SIGCHLD, childsig);
    signal(SIGWINCH, winchsig);
}

void winchsig()
{
    attention |= ATTN_WINCHSIG;
}

void childsig()
{
    attention |= ATTN_CHILDSIG;
}

void pipesig()
{
    exitcancan("\n#broken pipe.\n");
}

void termsig()
{
    exitcancan("\n#termination signal.\n");
}

void intrsig()
{
    if(sockfd > 1) close(sockfd);
    exitcancan("\n#interrupted.\n");
}

/*
 * suspend ourselves
 */
void suspend_cancan()
{
    if(signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	lastline();		/* maybe skip_edline() instead */
	reset_terminal();
	fflush(stdout);
	signal(SIGTERM, SIG_IGN);
	signal(SIGTSTP, SIG_DFL);
	kill(getpid(), SIGTSTP);
	set_terminal();
	signal(SIGTERM, termsig);
	check_winsize();	/* in case size changed meanwhile */
	lastline();		/* so we know where the cursor is */
	redraw_all();
    } else
	to_footnote("#your shell probably lacks job control.");
}

/*
 * terminate cancan as cleanly as possible
 */
void exitcancan(message)
char *message;
{
    lastline();
    reset_terminal();
    printf("%s#cancan stopped dancing.\n", message);
    exit(0);
}

/*
 * print an error message
 * return 1, so we can exit(syswarn())
 */
int syswarn(msg)
char *msg;
{
    int err = errno;
    if(err > 0 && err < sys_nerr)
	printf("%s (%s)\n", msg, sys_errlist[err]);
    else
	printf("%s\n", msg);
    return 1;
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
    /* NOTE: we can't be called before set_terminal(). hmm */
    int err = errno;
    reset_terminal();
    fprintf(stderr, "\n#fatal system call error: %s (", msg);
    if(err > 0 && err < sys_nerr)
	fprintf(stderr, "%s)\n", sys_errlist[err]);
    else
	fprintf(stderr, "%d)\n", err);
    exit(1);
}

/*
 * read definitions from file
 */
static void read_settings(f)
FILE *f;
{
    typedef enum {
	def_action, def_alias, def_bind, def_mark, def_hilite,
	def_option, def_savefile, def_init, def_host, def_error
    } def_token;
    static char * const def_names[] = {
	"#action", "#alias", "#bind", "#mark", "#hilite",
	"#option", "#savefile", "#init", "#host", 0
    };
    char buf[LINELEN];
    char *line, *args, *rest;
    int n_err = 0, ln = 0;
    def_token what;

    while(++ln, fgets(buf, sizeof(buf), f)) {
	if(*buf == '\n') continue;
	new_settings = 0;
	chop_str(line = buf, '\n');
	what = (def_token)get_token(&line, def_names);
	if(!(args = line))
	    what = def_error;

	switch(what) {
	case def_action:
	    parse_action(args); break;
	case def_alias:
	    parse_alias(args); break;
	case def_bind:
	    parse_bind(args); break;
	case def_mark:
	    parse_marker(args); break;
	case def_hilite:
	    parse_hilite(args); break;
	case def_option:
	    parse_options(args); break;
	case def_savefile:
	    if(strcmp(args, SAVEFILEVER)) {
		printf("Warning: expect definition file version '%s'\n",
			SAVEFILEVER);
		what = def_error;
	    }
	    break;
	case def_init:
	    str_assign(&initstr, args);
	    break;
	case def_host:
	    rest = chop_str(args, ' ');
	    if(rest && (portnumber = atoi(rest))) {
		str_assign(&hostname, args);
		break;
	    }
	    /* FALLTHROUGH */
	default:
	    printf("Bad definition file format, line %d\n", ln);
	    what = def_error;
	}

	if(what < def_option && !new_settings) {
	    /* NOTE: we don't catch redefinitions this way, though */
	    printf("Bad %s definition, line %d\n", def_names[what], ln);
	} else if(what != def_error)
	    continue;

	if(++n_err >= 10) {
	    /* printf("Too many errors encountered, quitting.\n"); */
	    exit(1);
	}
    }
    if(n_err) {
	/* printf("Errors were encountered; autosave turned off.\n"); */ 
	OPTION(OPT_AUTOSAVE) = 0;
    }
}

/*
 * save settings in definition file
 */
void save_settings()
{
    char buf[LINELEN], *cp;
    FILE *f;
    aliasnode *al;
    actionnode *ac;
    marknode *m;
    keynode *k;

    if(!(cp = get_env("SAVEFILE"))) return;
    if(!(f = fopen(cp, "w"))) {
	printf("#error: can't write to definition file %s\n", cp);
	return;
    }
    fprintf(f, "#savefile %s\n", SAVEFILEVER);
    print_options(buf);
    fprintf(f, "%s\n", buf);
    if(edattr != NOATTRCODE)
	fprintf(f, "#hilite %s\n", attr_name(edattr));
    if(hostname)
	fprintf(f, "#host %s %d\n", hostname, portnumber);
    if(initstr)
	fprintf(f, "#init %s\n", initstr);
    for(al = aliases; al; al = al->next)
	fprintf(f, "#alias %s=%s\n", al->name, al->subst);
    for(ac = actions; ac; ac = ac->next) {
	escape_specials(ac->pattern, buf, "\\\"");
	fprintf(f, "#action %c%s \"%s\"=%s\n", (ac->active < 0) ? '-' : '+',
		ac->label, buf, ac->commands);
    }
    for(m = markers; m; m = m->next) {
	escape_specials(m->pattern, buf, "\\= ");
	fprintf(f, "#mark %s=%s\n", buf, attr_name(m->attrcode));
    }
    /* NOTE: deleted default keys should be saved too, as "#bind <key>=" */
    for(k = keydefs; k; k = k->next)
	if(k->call_data) {
	    escape_seq(k->sequence, buf);
	    fprintf(f, "#bind %s %s=%s\n", k->name, buf, k->call_data);
	}
    fclose(f);
}

void to_footnote(msg)
char *msg;
{
    attention |= ATTN_FOOTNOTE;
    str_assign(&footnote, msg);
}

void to_capfile(str, nl)
char *str; int nl;
{
    /* assert(capfile); */
    fputs(str, capfile);
    if(nl) putc('\n', capfile);
}

char *str_dup(s)
char *s;
{
    char *k = (char *)malloc(strlen(s) + 1);
    if(!k) syserr("malloc");
    return strcpy(k, s);
}

void str_assign(str, s)
char **str, *s;
{
    char *new = s ? str_dup(s) : 0;
    if(*str) free(*str);
    *str = new;
}

/*
 * return the hash value of a string
 */
static char hash_var(str)
char *str;
{
    char *s = str, x = 0, c;
#if 0
    while((c = *s) && c != '=')
	s++, x = 31 * x + c;
#else
    char z = 0;
    while((c = *s) && c != '=')
	s++, x ^= z += c;
#endif
    return x;
}

/*
 * initialize our variables
 */
static void init_env()
{
    char **new_env, *new_hash;
    char **oldenv = environ;
    int i = 0;
    while(oldenv[i++]);
    new_env = (char **)malloc(i * sizeof(char *));
    new_hash = (char *)malloc(i);
    /* we are called before set_terminal -- can't syserr */
    if(!new_env || !new_hash)
	exit(syswarn("Unable to allocate own environment"));
    envtop = --i;
    new_env[i] = 0;
    while(--i >= 0)
	new_hash[i] = hash_var(new_env[i] = str_dup(oldenv[i]));
    environ = new_env;
    envhash = new_hash;
}

/*
 * look up a variable in the environ array
 * return the index to it, or -1 if not found
 */
static int find_env(var, len)
char *var;
int len;
{
    int i = envtop;
    char *s, c = hash_var(var);
    while(--i >= 0) {
	while(envhash[i] != c && --i >= 0);
	if(i < 0) break;
	if(!strncmp(s = environ[i], var, len))
	    if(!s[len] || s[len] == '=') break;
    }
    return i;
}

/*
 * analogous to getenv, but with fast lookup
 */
char *get_env(variable)
char *variable;
{
    char *s = variable;
    int len = strlen(variable);
    int i = find_env(s, len);
    if(i < 0) return 0;
    s = environ[i] + len;
    return *s ? s + 1 : s;
}

/*
 * set or unset (when value == NULL) an environment variable
 */
void set_env(variable, value)
char *variable, *value;
{
    int len = strlen(variable);
    int i = find_env(variable, len);
    if(value) {
	if(i < 0) {
	    i = ++envtop;
	    if(i >= envsize) {
		int new_size = (i | 63) + 1;
		envsize = new_size;
		environ = (char **)realloc(environ, new_size * sizeof(char *));
		if(!environ || !(envhash = (char *)realloc(envhash, new_size)))
		    syserr("realloc");
	    }
	    /* the env array needs to be zero terminated when we execve() */
	    environ[i] = 0;
	    envhash[--i] = hash_var(variable);
	}
	{
	    char buf[LINELEN], *d = buf;
	    strcpy(d, variable);
	    d += len;  *d++ = '=';
	    strcpy(d, value);
	    str_assign(environ + i, buf);
	}
    } else if(i >= 0) {
	int k;
	free(environ[i]);
	k = --envtop;
	if(k != i) {
	    envhash[i] = envhash[k];
	    environ[i] = environ[k];
	}
	environ[k] = 0;
    }
}

#ifndef OMIT_SESSIONS
/*
 * Below are multiple-session support functions.
 *
 * NOTE: sess_active[0] holds total no. of connections
 */

/*
 * return TRUE if current session being processed is main connection
 */
int is_main_sess()
{
    return (sockfd == main_sockfd);
}

/*
 * return session's index or socket descriptor given id,
 * or 0 if invalid id is given
 */
int find_sess(id, notify)
char *id;
int notify;
{
    int i, n;
    for(i = main_sockfd, n = sess_active[0]; n; i++)
	if(sess_id[i]) {
	    n--;
	    if(!strcmp(sess_id[i], id))
		return i;
	}
    if(notify)
	printf("#no such connection\n");
    return 0;
}

/*
 * connect another session
 */
int connect_sess(id, init, host, port)
char *id;	/* session id - NULL if main session */
char *init;	/* optional init string */
char *host;
int port;
{
    int fd;
    if((!host && !(host = hostname)) || (!port && !(port = portnumber))) {
	printf("#please specify both hostname and port number.\n");
	return -1;
    }
    /* dial the number by moving the right index in small circles */
    if((fd = doconnect(host, port)) < 0)
	return fd;
    FD_SET(fd, &fdset);		/* add socket to select() set */
    sess_active[0]++;
    if(!main_sockfd)		/* save main socket fd */
	main_sockfd = sockfd = fd;
    if(!id) {			/* main session */
	/*
	 * NOTE: make it "sess_active[fd] = 2;" if you think
	 * #snoop command should not deactivate main session.
	 */
	sess_active[fd] = 1;
	sess_id[fd] = "main";
    } else {
	sess_active[fd] = !init;
	sess_id[fd] = str_dup(id);
    }
    if(init) parse(init, 0, (struct param *)0);
    return fd;
}

/*
 * disconnect session
 */
void disconnect_sess(sfd)
int sfd;
{
    /* shutdown(sfd, 2); */
    close(sfd);
    printf("#connection \"%s\" closed.\n", sess_id[sfd]);
#if 0
    if(linemode & LM_CHAR) printf("\r");
#endif
    if(sfd == main_sockfd)		/* main connection closed */
	exitcancan("");
    FD_CLR(sfd, &fdset);
    sess_active[0]--;
    sess_active[sfd] = 0;
    str_assign(sess_id + sfd, (char *)0);
}

/*
 * toggle output display from another session
 */
void snoop_sess(id)
char *id;
{
    int sfd = find_sess(id, 1);
    if(sfd)
	sess_active[sfd] ^= 1;
}

/*
 * show open connections
 */
void show_connections()
{
    int i, n = sess_active[0];
    if(n) {
	sprintf(insertbuf, "#The following connection%s open:",
		(n == 1) ? " is" : "s are");
	for(i = main_sockfd; to_pager(insertbuf), --n >= 0; i++) {
	    while(!sess_id[i]) i++;
	    sprintf(insertbuf, "##%-8s%s", sess_id[i],
		    !sess_active[i] ? "" :
		    i == main_sockfd ?  " (active)" : " (snooped)");
	}
    } else
	printf("#No open connections.\n");
}
#endif
