/*
 * tel.c - telnet protocol communication module for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/telnet.h>
#ifndef TELOPT_NAWS
#define TELOPT_NAWS 31
#endif
#include <arpa/inet.h>
#ifndef NEXT
#include <unistd.h>
#endif

#ifdef TERM
#include <client.h>
#endif /* TERM */

#include "cancan.h"
#include "term.h"

int socket();
int connect();

#ifdef TELOPTS
#define TELOPTSTR(n) ((n) > NTELOPTS ? "unknown" : telopts[n])
#endif

static void sendopt();
static void dosubopt();
static void send_raw();

static int do_naws;		/* negotiate (send our) window size */

#ifndef OMIT_SESSIONS
static int output_socket;	/* to which socket buffer should be sent */
#else
#define output_socket (sockfd)
#endif

#define MAXSUBOPT	20	/* currently, this is more than sufficient */

static int output_len;		/* number of characters in output_buffer */
static char output_buffer[BUFSIZE];

/*
 * connect to remote host
 * NOTE: We block while connection is attempted. Handling sockets in
 * nonblocking mode is not too complicated. Host lookup... is quite
 * a different story.
 */
int doconnect(addr, port)
char *addr; int port;
{
    int fd;
    set_spec_keys(1);	/* we might hang here for a long time... */

#ifndef TERM
    do {
	struct sockaddr_in address;
	struct hostent *host_info;
	printf("Trying %s... ", addr); fflush(stdout);
	address.sin_family = AF_INET;
#if 0	/* inet_aton() is not quite portable */
	if(!inet_aton(addr, &address.sin_addr))
#else
	/*
	 * NOTE: inet_addr is documented to return -1 for malformed requests,
	 * but it is declared to return unsigned long!  Anyway, this works.
	 */
	address.sin_addr.s_addr = inet_addr(addr);
	if((int)address.sin_addr.s_addr == -1)
#endif
	    if((host_info = gethostbyname(addr))) {
		memcpy((char*)&address.sin_addr,
			host_info->h_addr, host_info->h_length);
		address.sin_family = host_info->h_addrtype;
	    } else {
		printf("unknown host\n");
		fd = -1;
		break;
	    }
	address.sin_port = htons(port);

	if((fd = socket(address.sin_family, SOCK_STREAM, 0)) < 0) {
	    syswarn("error creating socket");
	    break;
	}
	if(fd >= MAXFDSCAN) {
	    printf("not connected (Too many open connections)\n");
	    close(fd);  fd = -1;
	    break;
	}

	for(;;) {
	    if(connect(fd, (struct sockaddr *)&address, sizeof(address)) < 0) {
		if(errno == EINTR) continue;
		syswarn("host not responding");
		close(fd);  fd = -1;
	    } else
		printf("connected!\n");
	    break;
	}
    } while(0);

#else /* term */

    if((fd = connect_server(0)) < 0)
	printf("Unable to connect to term server\n");
    else {
	send_command(fd, C_PORT, 0, "%s:%d", addr, port);
	printf("Connected to term server...\n");
#ifdef TERM_COMPRESS
	send_command(fd, C_COMPRESS, 1, "y");
#endif
	send_command(fd, C_DUMB, 1, 0);
    }
#endif /* term */

    set_spec_keys(0);
    return fd;
}

/*
 * read a maximum of size bytes from remote host
 * using the telnet protocol. return bytes read.
 */
int read_socket(buffer, size)
char *buffer; int size;
{
    /*
     * NOTE: could add a GOTSBTTYPE state to handle the only suboption
     * negotiation we currently support, and do without subopt buffer.
     * NOTE: we should process NUL's and CR's (cut them out when needed)
     * and recognize MPI at the same time we scan for telnet commands.
     * We would then be called from mainlupe(), and call process_input()
     * or mp_receive(). The buffer here should be static as well.
     */
    typedef enum {
	NORMAL = 0, GOTIAC, GOTWILL, GOTWONT, GOTDO, GOTDONT, GOTSB, GOTSBIAC
    } telnetstate;
#ifndef OMIT_SESSIONS
    static telnetstate sock_state[MAXFDSCAN];
    telnetstate state = sock_state[sockfd];
#else
    static telnetstate sock_state;
    telnetstate state = sock_state;
#endif
    static unsigned char subopt[MAXSUBOPT];
    static int subchars;
    int i;
    unsigned char *p, *s = (unsigned char *)buffer;

    if((i = read(sockfd, s, size)) <= 0) {
#ifndef OMIT_SESSIONS
	if(i < 0) syswarn("#warning: read from socket");
	sock_state[sockfd] = NORMAL;
#else
	if(i < 0) syserr("read from socket");
#endif
	return -1;
    }

    /* scan through the buffer and interpret telnet protocol escapes */
    for(p = s; --i >= 0; s++) {
	switch(state) {
	default:
	case NORMAL:
	    while(*s == IAC
		    ? (state = GOTIAC, 0)
		    : (*p++ = *s++, --i >= 0));
	    break;

	case GOTIAC:
	    switch(*s) {
	    case WILL:
		state = GOTWILL; break;
	    case WONT:
		state = GOTWONT; break;
	    case DO:
		state = GOTDO; break;
	    case DONT:
		state = GOTDONT; break;
	    case SB:
		state = GOTSB;
		if(is_main_sess())	/* subopts ignored in subsession */
		    subchars = 0;
		break;
	    case IAC:
		*p++ = IAC;
		state = NORMAL;
		break;
		/*
	    case GA:
		 * NOTE: GA could be handled as end-of-prompt marker one day.
		 * However, prompts have to be identified as such in the
		 * absence of GA anyway, and what is there to gain after all?
		 */
	    default:
		/* ignore the rest of the telnet commands */
#ifdef TELOPTS
		fprintf(stderr, "[skipped IAC <%d>]\n", *s);
#endif
		state = NORMAL;
		break;
	    }
	    break;

	/*
	 * NOTE: RFC 854 says we should not acknowledge "requests to enter
	 * some mode we are already in", to prevent negotiation loops. hmm.
	 *
	 * NOTE: perhaps we should handle some more telnet options, like:
	 * - TELOPT_BINARY: MP transfers should probably be carried out
	 * in binary mode, and MPI recognized only after this option.
	 * - TELOPT_LINEMODE: the proper way to switch linemode. However,
	 * it seems unnecessarily complex for our purposes, and the klugly
	 * SGA option probably needs to be processed anyway.
	 * - TELOPT_RCP: rarely used, why bother?
	 */

	/*
	 */

	case GOTWILL:
#ifdef TELOPTS
	    fprintf(stderr, "[got WILL %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	    case TELOPT_ECHO:
		/*
		 * host echoes, turn off echo here
		 * we don't want subsidiary session password entries
		 * to block anything in the main session, though.
		 */
		if(is_main_sess())
		    linemode |= LM_NOECHO;
		sendopt(DO, *s);
		break;
	    case TELOPT_SGA:		/* this can't hurt */
		if(is_main_sess()) {
		    linemode |= LM_CHAR;
		    set_spec_keys(1);
		}
		sendopt(DO, *s);
		break;
	    default:
		/* don't accept other options */
		sendopt(DONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;

	case GOTWONT:
#ifdef TELOPTS
	    fprintf(stderr, "[got WONT %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_ECHO) {
		/* host no longer echoes, we do it instead */
		if(is_main_sess())
		    linemode &= ~LM_NOECHO;
	    }
	    /* case TELOPT_SGA ? */
	    /* accept any WONT */
	    sendopt(DONT, *s);
	    state = NORMAL;
	    break;

	case GOTDO:
#ifdef TELOPTS
	    fprintf(stderr, "[got DO %s]\n", TELOPTSTR(*s));
#endif
#ifndef OMIT_SESSIONS
	    if(!is_main_sess()) {
		sendopt(WONT, *s);
	    } else
#endif
	    switch(*s) {
	    case TELOPT_SGA:
		linemode |= LM_CHAR;
		set_spec_keys(1);
		/* FALLTHROUGH */
	    case TELOPT_TTYPE:
		sendopt(WILL, *s);
		break;
	    case TELOPT_NAWS:
		sendopt(WILL, *s);
		do_naws = 1;
		send_term_size();
		break;
		/* FALLTHROUGH */
	    default:
		/* accept nothing else */
		sendopt(WONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;

	case GOTDONT:
#ifdef TELOPTS
	    fprintf(stderr, "[got DONT %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	    case TELOPT_SGA:
		if(is_main_sess()) {
		    linemode &= ~LM_CHAR;
		    set_spec_keys(0);
		}
		break;
	    case TELOPT_NAWS:
		do_naws = 0;
		break;
	    }
	    sendopt(WONT, *s);
	    state = NORMAL;
	    break;

	case GOTSB:
	    if(*s == IAC) {
		state = GOTSBIAC;
	    } else if(is_main_sess()) {	/* subopts ignored in subsession */
		if(subchars < MAXSUBOPT)
		    subopt[subchars++] = *s;
	    }
	    break;

	case GOTSBIAC:
	    if(*s == IAC) {
		state = GOTSB;
		if(is_main_sess()) {	/* subopts ignored in subsession */
		    if(subchars < MAXSUBOPT)
			subopt[subchars++] = *s;
		}
	    } else if(*s == SE) {
		if(is_main_sess())	/* subopts ignored in subsession */
		    dosubopt((char*)subopt, subchars);
		state = NORMAL;
	    } else {
		/*
		 * NOTE: telnet would probably try to process the subopt here.
		 * we just ignore it and hope it goes away.
		 */
		to_footnote("#warning: telnet suboption not IAC SE terminated");
		state = NORMAL;
	    }
	    break;
	}
    }
#ifndef OMIT_SESSIONS
    sock_state[sockfd] = state;
#else
    sock_state = state;
#endif
    return (char*)p - buffer;
}

/*
 * send an option negotiation
 * 'what' is one of WILL, WONT, DO, DONT
 */
static void sendopt(what, opt)
int what; unsigned opt;
{
    static unsigned char buf[3] = { IAC, 0, 0 };
    buf[1] = what; buf[2] = opt;
    send_raw(buf, 3);
#ifdef TELOPTS
    fprintf(stderr, "[sent %s %s]\n", (what == WILL) ? "WILL" :
	   (what == WONT) ? "WONT" :
	   (what == DO) ? "DO" : (what == DONT) ? "DONT" : "error",
	   TELOPTSTR(opt));
#endif
}

/*
 * Send current terminal size (RFC 1073)
 */
void send_term_size()
{
    if(do_naws) {
	static unsigned char buf[9] =
		{ IAC, SB, TELOPT_NAWS, 0, 0, 0, 0, IAC, SE };
	buf[3] = cols >> 8;
	buf[4] = cols & 0xff;
	buf[5] = lines >> 8;
	buf[6] = lines & 0xff;
	send_raw(buf, 3);
	/* buf[3..6] are data, and need to be escaped if equal to IAC */
	send_to_host(buf + 3, 4);
	send_raw(buf + 7, 2);
#ifdef TELOPTS
	fprintf(stderr, "[term size %d %d]\n", cols, lines);
#endif
    }
}

/*
 * process suboptions.
 * right now, only terminal type is processed.
 */
static void dosubopt(str, len)
char *str;
int len;
{
    if(len < 2) return;
    if(str[0] == TELOPT_TTYPE && str[1] == 1) {		/* 1 == SEND */
	char buf[64], *term;
	/*
	 * NOTE: we do not verify that we are sending a valid terminal name.
	 * Perhaps we should record our term type at termcap init, though.
	 * NOTE: should we ignore TERM variable if VT100 is defined?
	 */
	if(!(term = get_env("TERM"))) term = "unknown";
#if defined(_POSIX_SOURCE)
	send_raw(buf, sprintf(buf, "%c%c%c%c%.40s%c%c",
		IAC, SB, TELOPT_TTYPE, 0, term, IAC, SE));	/* 0 == IS */
#else
	sprintf(buf, "%c%c%c%c%.40s%c%c",
		IAC, SB, TELOPT_TTYPE, 0, term, IAC, SE);	/* 0 == IS */
	send_raw(buf, 4 + strlen(buf + 4));
#endif
#ifdef TELOPTS
	fprintf(stderr, "[got SB TERMINAL TYPE SEND]\n");
	fprintf(stderr, "[sent SB TERMINAL TYPE IS %.40s]\n", term);
#endif
    }
}

/*
 * put data in output buffer as is (don't escape IAC's).
 * assert(len <= sizeof(output_buffer)
 */
static void send_raw(data, len)
char *data; int len;
{
    if(sockfd != output_socket || output_len + len > sizeof(output_buffer)) {
	flush_output_buffer();
	output_socket = sockfd;
    }
    memcpy(output_buffer + output_len, data, len);
    output_len += len;
}

/*
 * put data in the output buffer for transmission to the remote host
 */
void send_to_host(data, len)
char *data; int len;
{
    char *s;
    int i, n;
#ifndef OMIT_SESSIONS
    if(output_socket != sockfd) {	/* is there data to another socket? */
	flush_output_buffer();		/* then flush it */
	output_socket = sockfd;
    }
#endif
    for(s = data, i = output_len, n = len; n; n--) {
	if(i >= sizeof(output_buffer) - 1) {
	    output_len = i;
	    flush_output_buffer();
	    i = 0;
	}
	if((output_buffer[i++] = *s++) == (char)IAC)
	    output_buffer[i++] = (char)IAC;
    }
    output_len = i;
}

/*
 * send all buffered data to the remote host
 */
void flush_output_buffer()
{
    if(output_len) {
	/* NOTE: no guarantee we don't block here indefinitely. */
	if(output_socket)
	    while(write(output_socket, output_buffer, output_len) < 0)
		if(errno != EINTR) { syserr("write to socket"); break; }
	output_len = 0;
    }
}
