/*
 * term.c - terminal handling routines for cancan
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <sys/ioctl.h>

#ifndef USESGTTY
#ifdef APOLLO
#include "/sys5.3/usr/include/sys/termio.h"
#else
#ifdef AIX
#include <termio.h>
#else
#include <termios.h>
#endif
#endif
/*
#else
#include <bsd/sgtty.h>
 */
#endif /* USESGTTY */

#ifdef BSD_386
#include <sys/ioctl_compat.h>
#define O_RAW RAW
#define O_ECHO ECHO
#define O_CRMOD CRMOD
#define O_CBREAK CBREAK
#endif

#ifndef USESGTTY
/*
 * NOTE: we should probably use TCSETSW/TCSETAW/TIOCSETAW instead,
 * since we could be toggling the OPOST flag...
 * Or we could use tcsetattr/tcgetattr rather than direct ioctl.
 */
#if defined(TCSETS) || defined(TCSETATTR)
#ifndef TCSETS		/* HP-UX defines TCSETATTR/TCGETATTR? */
#define TCSETS TCSETATTR
#define TCGETS TCGETATTR
#endif
typedef struct termios termiostruct;
#else
/* if TCSETA is not defined either, we could try to use TIOCSETA */
#define TCSETS TCSETA
#define TCGETS TCGETA
typedef struct termio termiostruct;
#endif
#endif /* !USESGTTY */

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"

static void gotoxy();
static void optmove();
static int wrap_edline();
static int edline_feed();
static void put_s();
static void put_sn();
static void hilite();
static int pm_match();
static void pm_preprocess();
static void mark_and_print();
static char *wrap_pager_line();
static int pager_prompt();

#ifndef VT100

int tgetent();
int tgetnum();
int tgetflag();
char *tgetstr();
char *tgoto();

static int extract();

/* terminal escape sequences */
char kpadstart[CAPLEN], kpadend[CAPLEN];
char clreoln[CAPLEN], clreoscr[CAPLEN];
char begoln[CAPLEN], curgoto[CAPLEN];
char leftcur[CAPLEN], upcur[CAPLEN];
char insstart[CAPLEN], insstop[CAPLEN];
char inschar[CAPLEN], delchar[CAPLEN];
/* attribute changers: */
char modebold[CAPLEN], modeuline[CAPLEN], modeinv[CAPLEN];
char modestandon[CAPLEN], modestandoff[CAPLEN];
char modenorm[CAPLEN];

char key_left[CAPLEN], key_right[CAPLEN];
char key_up[CAPLEN], key_down[CAPLEN];

static int len_begoln, len_clreoln;
static int len_leftcur, len_upcur, gotocost;
static int insertcost, insertfinish;
static int insert1_cost, deletecost;

#if !defined(XASCII)
int metakey;		/* terminal has meta key setting high bit */
#endif
int wrapglitch;		/* strange wrap behaviour (a la vt100) */

#else	/* hard-coded vt100 features if no termcap: */

char clreoln[] = "\033[K", clreoscr[] = "\033[J";
char begoln[] = "\r", curgoto[] = "\033[%d;%dH";
char leftcur[] = "\b", upcur[] = "\033[A";

char modestandon[] = "", modestandoff[] = "";
char modebold[] = "\033[1m", modeuline[] = "\033[4m", modeinv[] = "\033[7m";
char modenorm[CAPLEN] = "\033[m";

char key_left[] = "\033[D", key_right[] = "\033[C";
char key_up[] = "\033[A", key_down[] = "\033[B";

#define len_begoln	1
#define len_leftcur	1
#define len_upcur	3
#define len_clreoln	3
#define gotocost	8

#define wrapglitch	1
#define insertfinish	0

#endif /* VT100 */

#ifndef LINES
#define LINES 24
#endif
#ifndef COLUMNS
#define COLUMNS 80
#endif

int lines = LINES;	/* screen size */
int cols = COLUMNS;

int lines_1 = LINES - 1;/* last screen line */
int cols_1 = COLUMNS;	/* either cols, or cols-1 if no wrapglitch */

int line0 = 0;		/* screen line where the input line starts */
int col0 = 0;		/* input line offset (= length of prompt) */

/* input line highlighting */
int edattr = NOATTRCODE;/* input line attributes */
char edattrbeg[CAPLEN];	/* start input line attributes */
char edattrend[CAPLEN]; /* end input line attributes */

static int pager_lines;	/* num lines pager has currently displayed */

#ifdef XTERM_JS_BUG
static int leave_bg_off;
#endif

#ifndef USESGTTY
static termiostruct ttybsave;
#else
static struct sgttyb ttybsave;
static struct tchars tcsave;
static struct ltchars ltcsave;
#endif

/*
 * Save the original state, disable echo and special keys.
 */
void set_terminal()
{
#ifndef USESGTTY
    ioctl(0, TCGETS, &ttybsave);
#else
    ioctl(0, TIOCGETP, &ttybsave);
    ioctl(0, TIOCGETC, &tcsave);
    ioctl(0, TIOCGLTC, &ltcsave);
#endif
    set_spec_keys(0);
    signal(SIGINT, intrsig);
#ifndef VT100
    put_s(kpadstart);
#endif
}

/*
 * Reset the terminal to its original state
 */
void reset_terminal()
{
    signal(SIGINT, SIG_IGN);
#ifndef USESGTTY
    ioctl(0, TCSETS, &ttybsave);
#else
    ioctl(0, TIOCSETP, &ttybsave);
    ioctl(0, TIOCSETC, &tcsave);
    ioctl(0, TIOCSLTC, &ltcsave);
#endif
    hilite(0);
#ifndef VT100
    put_s(kpadend);
#endif
}

/*
 * set raw/cooked mode depending on the current linemode
 * enable/disable special keys
 */
void set_spec_keys(k)
int k;
{
    /*
     * NOTE: we should remember to reset the prompt upon
     * returning to line mode somewhere.
     */
#ifndef USESGTTY
    termiostruct ttyb;
    ttyb = ttybsave;
    ttyb.c_cc[VTIME] = 0;
    ttyb.c_cc[VMIN] = 1;
    ttyb.c_oflag |= OPOST;
    if(linemode & LM_CHAR) ttyb.c_oflag &= ~OPOST;
    ttyb.c_lflag &= ~(ECHO | ISIG | ICANON | IEXTEN);
    if(k) ttyb.c_lflag |= ISIG;
    ioctl(0, TCSETS, &ttyb);
#else /* USESGTTY */
    static const struct tchars tc_off = { -1, -1, -1, -1, -1, -1 };
    static const struct ltchars ltc_off = { -1, -1, -1, -1, -1, -1 };
    const struct tchars *tc = &tc_off;
    const struct ltchars *ltc = &ltc_off;
    struct sgttyb ttyb;
    ttyb = ttybsave;
    ttyb.sg_flags &= ~(O_ECHO | O_RAW);
    ttyb.sg_flags |= O_CBREAK | O_CRMOD;
    if(linemode & LM_CHAR) ttyb.sg_flags ^= O_RAW | O_CRMOD;
    ioctl(0, TIOCSETP, &ttyb);
    if(k) tc = &tcsave, ltc = &ltcsave;
    ioctl(0, TIOCSETC, tc);
    ioctl(0, TIOCSLTC, ltc);
#endif
}

#if !defined(VT100) || defined(TELNETBUG)
/*
 * read termcap definitions
 */
void settermcap()
{
    char *term = get_env("TERM");

#ifndef VT100
    /*
     * NOTE: We should check if term has "in" insert_null_glitch (rare).
     * And use capabilities which take a parameter ("IC", "DC", ...)!
     */
    struct tc_init_node {
	char cap[3], critical;
	char *buf;
	int *len;
    };
    static const struct tc_init_node tc_init[] = {
	{ "cm", 1, curgoto, 0 },
	{ "ce", 1, clreoln, &len_clreoln },
	{ "cd", 1, clreoscr, 0 },
	{ "le", 0, leftcur, &len_leftcur },
	{ "up", 0, upcur, &len_upcur },
	{ "cr", 0, begoln, &len_begoln },
	{ "im", 0, insstart, &insertcost },
#ifndef TERMPCAP_ORTHODOX
	{ "ic", 0, inschar, &insert1_cost },
#else
	{ "ic", 0, insstart, &insertcost },
#endif
	{ "ei", 0, insstop, &insertcost },
	{ "dm", 0, delchar, &deletecost },
	{ "dc", 0, delchar, &deletecost },
	{ "ed", 0, delchar, &deletecost },
	{ "me", 0, modenorm, 0 },
	{ "md", 0, modebold, 0 },
	{ "mr", 0, modeinv, 0 },
	{ "us", 0, modeuline, 0 },
	{ "so", 0, modestandon, 0 },
	{ "se", 0, modestandoff, 0 },
	{ "ks", 0, kpadstart, 0 },
	{ "ke", 0, kpadend, 0 },
	{ "kl", 0, key_left, 0 },
	{ "kr", 0, key_right, 0 },
	{ "ku", 0, key_up, 0 },
	{ "kd", 0, key_down, 0 },
	{ "",   0, 0, 0 }
    };
    const struct tc_init_node *np;
    char tcbuf[2048];		/* by convention, this is enough */
    int i;
    if(!term || !(i = tgetent(tcbuf, term))) {
	if(term) fprintf(stderr,
		    "There is no entry for '%s' in the terminal data base.\n",
		    term);
	fprintf(stderr, "Please set your TERM environment variable.\n");
	exit(1);
    }
    if(i < 0)	/* NOTE: perhaps we should assume default vt100 values? */
	exit(syswarn("Couldn't read termcap database"));
    for(np = tc_init; np->cap[0]; np++)
	if((i = extract(np->cap, np->buf))) {
	    if(np->len) *np->len += i;
	} else if(np->critical) {
	    fprintf(stderr,
		    "Your terminal is not powerful enough, missing '%s'.\n",
		    np->cap);
	    exit(1);
	}
    if(!len_begoln)
	strcpy(begoln, "\r"), len_begoln = 1;
    if(!len_leftcur)
	strcpy(leftcur, "\b"), len_leftcur = 1;
    /* check_winsize() will override these later */
    if((i = tgetnum("li")) > 0) lines_1 = (lines = i) - 1;
    if((i = tgetnum("co")) > 1) cols_1 = cols = i;
    gotocost = strlen(tgoto(curgoto, cols_1, lines_1));
#if !defined(XASCII)
    metakey = tgetflag("km");
#endif
    wrapglitch = tgetflag("xn");
    if(!wrapglitch)
	cols_1--, insertfinish = gotocost + len_clreoln;
#ifndef TERMCAP_ORTHODOX
    if(insert1_cost && (!insertcost || insertcost > insert1_cost)) {
	insertcost = insert1_cost;
	strcpy(insstart, inschar);
	insstop[0] = 0;
    }
#endif
#endif /* ! VT100 */

#ifdef TELNETBUG
    if(!term || !strncmp(term, "vt10", 4)) {
	/* might be NCSA Telnet 2.2 for PC, which doesn't reset colours */
	sprintf(modenorm, "\033[;3%d;4%dm", DEFAULTFG, DEFAULTBG);
    }
#endif
}
#endif /* !defined(VT100) || defined(TELNETBUG) */

#ifndef VT100
/*
 * extract termcap 'cap' and strcat it to buf.
 * return the lenght of the extracted string.
 */
static int extract(cap, buf)
char *cap, *buf;
{
    char *d = buf += strlen(buf);
    char *s = tgetstr(cap, &buf);
    int len = 0;
    /*
     * NOTE: Padding information is removed. We assume that no terminals
     * need padding nowadays. At least it makes things much easier.
     * NOTE: We don't check for buf overflow.
     */
    if(s) {
	s += strspn(s, "0123456789*");
	for( ; *s; *d++ = *s++, len++)
	    if(*s == '$' && *(s + 1) == '<')
		if(!(s = strchr(s, '>')) || !*++s) break;
    }
    *d = 0;
    return len;
}
#endif /* VT100 */

/*
 * add the default keypad bindings to the list
 */
void keypad_bindings()
{
    /*
     * NOTE: termcap doesn't have sequences for the numeric keypad,
     * so we just assume they are the same as for a vt100.
     */
    struct keypad_init_node {
	char label[4], seq[4], cmd[6];
    };
    static const struct keypad_init_node keypad_init[] = {
	{ "kp2", "\033Or", "s" },
	{ "kp3", "\033Os", "d" },
	{ "kp4", "\033Ot", "w" },
	{ "kp5", "\033Ou", "exits" },
	{ "kp6", "\033Ov", "e" },
	{ "kp8", "\033Ox", "n" },
	{ "kp9", "\033Oy", "u" },
	{ "", "", "" }
    };
    const struct keypad_init_node *p = keypad_init;
    do {
	add_keynode(p->label, p->seq, run_command, p->cmd);
    } while((++p)->label[0]);
}

/*
 * initialize the key binding list
 */
void default_bindings()
{
    /*
     * NOTE: no check is made to see if (up, down, etc.) keys have a valid
     * keysequence. could use key_collision(), which is defined in cmd.c
     * NOTE: perhaps we should add these nodes in reverse order (and not
     * sorted), since most frequently used keys come first in defbinds[].
     */
    const struct defbind_node *p = defbinds;
    do {
	if(p->seq[0]) add_keynode(p->label, p->seq, p->funct, (char *)0);
    } while((++p)->label[0]);
}

/*
 * position the cursor using absolute coordinates
 * NOTE: tgoto is quite a piece of code, willing to handle padding,
 * which we chose to ignore, etc., etc.
 */
static void gotoxy(col, line)
int col, line;
{
#ifdef VT100
  printf(curgoto, line + 1, col + 1);
#else
  put_s(tgoto(curgoto, col, line));
#endif
}

/*
 * optimized cursor movement
 * from (fromline, fromcol) to (toline, tocol)
 * if tocol > 0, (toline, tocol) must reside on editline.
 */
static void optmove(fromline, fromcol, toline, tocol)
int fromline, fromcol, toline, tocol;
{
    static char buf[CAPLEN];
    char *cp = buf;
    int i, cost = 0;
    /*
     * First, move vertically to the correct line, then horizontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (gotoxy), use that.
     */
    for(;;) {
	if((i = toline - fromline) < 0) {
	    if(!len_upcur || (cost = -i * len_upcur) >= gotocost) break;
	    do strcpy(cp, upcur), cp += len_upcur; while(++i);
	} else if(i) {
	    /* lf's are mapped into crlf on output */
	    if((cost = 2 * i) >= gotocost) break;
	    do *cp++ = '\n'; while(--i);
	    fromcol = 0;
	}
	if(fromcol >= cols) break;	/* we are in a glitch state */
	if((i = tocol - fromcol) < 0) {
	    int dist = -i * len_leftcur;
	    if(dist <= len_begoln + tocol) {
		if(cost + dist > gotocost) break;
		do strcpy(cp, leftcur), cp += len_leftcur; while(++i);
	    } else {
		if((cost += len_begoln) >= gotocost) break;
		strcpy(cp, begoln);  cp += len_begoln;
		fromcol = 0;  i = tocol;
	    }
	}
	if(i) {
	    if(cost + i > gotocost) break;
	    if(fromcol < col0 && toline == line0) {
		/* gotoxy if hiliting in effect or special chars in prompt */
		if(edattrbeg[0] || promptstr[col0]) break;
		/*
		 * FEATURE: it should be recorded somewhere whether printed
		 * prompt was marked. In that case, gotoxy must be used
		 */
		strcpy(cp, promptstr);
		cp += fromcol = col0;
	    }
	    hilite(1);
	    memcpy(cp, edbuf + (toline - line0) * cols_1 + fromcol - col0,
		    tocol - fromcol);
	    cp += tocol - fromcol;
	}
	*cp = 0;
	put_s(buf);
	return;
    }
    gotoxy(tocol, toline);
}

/*
 * change the position on input line (optmove there, and set pos)
 */
void moveto(new_pos)
int new_pos;
{
    /*
     * FEATURE: the line we are moving to might be less than 0, or greater
     * than lines_1, if the display is too small to hold the whole editline.
     * In that case, the input line should be (partially) redrawn.
     */
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(new_pos), CURCOL(new_pos));
    pos = new_pos;
}

/*
 * move to the beginning of last line
 */
void lastline()
{
    /* NOTE: could use tcap "ll" here */
    gotoxy(0, line0 = lines_1);
}

/*
 * print a newline and advance line0.
 */
void newline()
{
    putc('\n', stdout);
    if(line0 < lines_1) line0++;
#ifdef XTERM_JS_BUG
    else leave_bg_off = 1;
#endif
}

/*
 * display or clear a one-line footnote at the bottom of the screen
 */
void display_footnote()
{
    char *msg = footnote;
    int to_c = CURCOL(pos);
    int to_l = CURLINE(pos);
    optmove(to_l, to_c, lines_1, 0);
    if(msg) {
	char *pat = "%s";
	hilite(0);
	if(CURLINE(edlen) >= lines_1)
	    --line0, --to_l, pat = "\n%s";
	printf(pat, msg);
	/* fflush(stdout); sleep(1); */
	gotoxy(to_c, to_l);
    } else {
	put_s(clreoln);
	optmove(lines_1, 0, to_l, to_c);
    }
}

/*
 * execute string as if typed
 */
void run_command(cmd)
char *cmd;
{
    clear_input();
    line0 = wrap_edline(cmd, strlen(cmd));
    hilite(0);  newline();
    if(capfile) to_capfile(cmd, 1);
    parse(cmd, 1, (struct param *)0);
    redraw_all();
}

/*
 * skip over the edit line, set new line0 and turn edattribs off
 */
void skip_edline()
{
    int p = pos;
    int cl = CURLINE(p);
    int nl = p - CURCOL(p);
    hilite(0);
    line0 = cl;
    do {
	newline();
	nl += cols_1;
    } while(nl < edlen);
}

/*
 * wrap out an [edit] line, leave cursor at the end of it
 */
static int wrap_edline(line, len)
char *line; int len;
{
    int p = 0;
    int cl = line0;
    int nl = -col0;
    for( ; ; ) {
	hilite(1);
	nl += cols_1;
	if(nl > len) break;
	put_sn(line + p, nl - p); p = nl;
	cl = edline_feed(cl);
    }
    put_sn(line + p, len - p);
    return cl;
}

/*
 * like newline(), but for use with edline operations
 */
static int edline_feed(ln)
int ln;
{
    if(++ln > lines_1) {
	ln = lines_1;
	--line0;
#ifndef NEW_BG_IS_DFLT
	hilite(-1);
#ifdef XTERM_JS_BUG
	leave_bg_off = 1;
#endif
#endif
    }
    put_s("\n");
    return ln;
}

/*
 * clear current input line (but not the prompt)
 * cursor is left right after the prompt
 */
void clear_input()
{
    if(pager_text) { 
	pager(0);
	return;
    }
    hilite(0);
    if(edlen) {
	int from_c = CURCOL(pos);
	int from_l = CURLINE(pos);
	if(line0 < lines_1) {
	    /* NOTE: plain clreoscr would suffice when col0 == 0 */
	    optmove(from_l, from_c, line0 + 1, 0);
	    put_s(clreoscr);
	    from_c = 0;  from_l = line0 + 1;
	}
	optmove(from_l, from_c, line0, col0);
	put_s(clreoln);
    }
}

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
void redraw_input()
{
    int len = edlen;
    if(len && !pager_text) {
	wrap_edline(edbuf, len);
	if(len != pos) {
	    int t = pos;
	    pos = len;
	    moveto(t);
	}
#ifdef XTERM_JS_BUG
	if(leave_bg_off) hilite(-1);
#endif
    }
}

/*
 * delete n characters at current position (the position is unchanged)
 * assert(n <= edlen - pos)
 */
void deletechars(n)
int n;
{
    /*
     * NOTE: the cost calculation here is not an exact science, but
     * that's ok. There is no excuse for not using "DC", however.
     */
    int p = pos;
    int nl, cl;		/* next line's starting pos, current line */
    int r_cost, d_cost;
    edlen -= n;
#if 0
    memmove(edbuf + p, edbuf + p + n, edlen - p);
    edbuf[edlen] = 0;
#else
    {
	int i, k = edlen;
	for(i = p; i < k; i++)
	    edbuf[i] = edbuf[i + n];
	edbuf[i] = 0;
    }
#endif
    nl = p - CURCOL(p);
    cl = CURLINE(p);
    d_cost = -gotocost;	/* no goto after delete in one-line operation */
    for( ; ; ) {
	d_cost -= len_clreoln;	/* delete is advantageous on last line */
	r_cost = edlen;
	nl += cols_1;
	if(edlen + n > nl) {	/* need process next line too */
	    d_cost = edlen - nl;
	    if(d_cost > 0)	/* new edline continues to next line */
		d_cost = 0, r_cost = nl;
	    d_cost += gotocost + n;
	}
	r_cost -= p;
	if(r_cost) {
#ifndef VT100
#ifdef XTERM_ICH_GLITCH
	  if(BACKGROUND(edattr) == ATTR_NONE)
#endif
	    if(deletecost && deletecost * n + d_cost < r_cost) {
#ifndef NEW_BG_IS_DFLT
		if(edlen + n <= nl) hilite(-1);
#endif
		for(d_cost = n; d_cost; d_cost--) put_s(delchar);
		if(edlen + n <= nl) break;
		gotoxy(cols_1 - n, cl);
		r_cost += p;  r_cost -= (p = nl - n);
	    }
#endif
	    hilite(1);
	    put_sn(edbuf + p, r_cost); p += r_cost;
	}
	d_cost = 0;
	if(p < nl) {
	    hilite(-1);
	    if(n == 1 && p < nl - 1) {
		p++; put_s(" ");
	    } else
		put_s(clreoln);
	    if(edlen + n <= nl) break;
	}
	put_s("\n"); cl++; p = nl;
	if(edlen <= nl) {
	    hilite(-1);
	    put_s(clreoscr);
	    break;
	}
    }
    if(p != pos) {
	int t = pos;
	pos = p;
	moveto(t);
    }
}

/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 * NOTE: we should optimize here for the usual (pos == edlen) case.
 */
void insertchars(str, n)
char *str; int n;
{
    int p = pos;
    int nl, cl;		/* next line's starting pos, current line */
#if 0
    memmove(edbuf + p + n, edbuf + p, edlen - p + 1);
    memcpy(edbuf + p, str, n);
#else
    {
	int i = n, k = edlen;
	do edbuf[i + k] = edbuf[k]; while(--k >= p);
	while(--i >= 0) edbuf[i + p] = str[i];
    }
#endif
    edlen += n;  pos += n;
    nl = p - CURCOL(p);
    cl = CURLINE(p);
    do {
	int i_cost = n;
	int r_cost = edlen;
	nl += cols_1;
	if(r_cost > nl)
	    r_cost = nl, i_cost += insertfinish;
	r_cost -= p;
#ifndef VT100
	/*
	 * NOTE: insert mode is used only when one char is inserted
	 * (which is probaly true > 90% of the time). Simplicity
	 */
	hilite(1);
	if(n == 1 && insertcost
#ifdef XTERM_ICH_GLITCH
		&& BACKGROUND(edattr) == ATTR_NONE
#endif
		&& insertcost + i_cost < r_cost) {
	    put_s(insstart);
	    put_sn(edbuf + p, 1); p++;
	    put_s(insstop);
	    if(insertfinish && edlen > nl) {
		gotoxy(cols_1, cl);
		put_s(clreoln);
	    }
	} else
#endif /* ! VT100 */
	{
	    put_sn(edbuf + p, r_cost);  p += r_cost;
	}
	if(edlen < nl) break;
	cl = edline_feed(cl); p = nl;
    } while(edlen > nl);
    if(p != pos) {
	int t = pos;
	pos = p;
	moveto(t);
    }
#ifdef XTERM_JS_BUG
    if(leave_bg_off) hilite(-1);
#endif
}

/*
 * print a char on current position (overwrite), advance position
 */
void overtype_char(c)
char c;
{
    hilite(1);
    edbuf[pos++] = c;
    putc(c, stdout);
    if(!CURCOL(pos))
	put_s("\n");
}

/*
 * print a (short) null terminated string
 */
static void put_s(str)
char *str;
{
    /* return fputs(str); */
    char c, *s;
    for(s = str; (c = *s); s++)
	putc(c, stdout);
}

/*
 * print a string with length n
 */
static void put_sn(str, len)
char *str; int len;
{
    fwrite(str, 1, len, stdout);
}

/*
 * NOTE: coloured background handling can be a pain.
 * We need to turn bg. off whenever new lines are created with scrolling, or
 * blanks shifted in when delchar is used. (New space is usually created with
 * current bg. not default bg.)  See README.xterm for description of some of
 * the coloured bg. related bugs in xterm. Old color_xterm, then again, has a
 * buggy cursor which taints the underlying text. The workaround for that is
 * too ugly to seriousely consider (i.e. redraw/spaceout each pos. where bg.
 * is changed).  In short, get yourself the latest color_xterm.
 */

/*
 * toggle input line attributes
 * Argument 0 means attributes OFF, 1 is attributes ON, and -1 means
 * we don't care as long as background if OFF.
 * NOTE: we should do attr ==> attr switching, not simply attr on/off. 
 */
static void hilite(k)
int k;
{
    static int in_effect;
    if(k < 0) {
#ifdef XTERM_JS_BUG
	leave_bg_off = 0;
#endif
	if(BACKGROUND(edattr) == ATTR_NONE) return;
	k = 0;
    }
    if(in_effect != k) {
	in_effect = k;
	put_s(k ? edattrbeg : edattrend);
    }
}

/*
 * put escape sequences to turn on/off an attribute into given buffers
 */
void attr_string(attrcode, begin, end)
int attrcode; char *begin, *end;
{
    attr_token fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);
    *begin = *end = '\0';
    if(fg < ATTR_COLOURS) {
	if(bg < ATTR_COLOURS)
	    sprintf(begin, "\033[3%d;4%dm", fg, bg);
	else
	    sprintf(begin, "\033[3%dm", fg);
    } else if(bg < ATTR_COLOURS) {
	sprintf(begin, "\033[4%dm", bg);
    } else {
	char *src;
	if(fg == ATTR_UNDERLINE)
	    src = modeuline;
	else if(fg == ATTR_BOLD)
	    src = modebold;
	else if(fg == ATTR_INVERSE || fg == ATTR_REVERSE)
	    src = modeinv;
	else return;
	if(!*src) {
	    strcpy(begin, modestandon);
	    strcpy(end, modestandoff);
	    return;
	}
	strcpy(begin, src);
    }
    strcpy(end, modenorm);
}

/*
 * NOTE: An altogether different pattern matching scheme should be devised--
 * probably one with marker info (and actions?) embedded in the pattern.
 *
 * Marking is performed regardless of possible attribute changes in the text.
 * Cancan uses attributes, but doesn't really handle them. For instance, try
 * "#mark \ =bold", and see what output is produced for long space sequences.
 */

static unsigned char delta[LINELEN], alpha[256];

/*
 * NOTE: in anticipation of trying several patterns against the text,
 * some preprocessing is done on the latter. (hashing, sort of).
 * alpha[c] is an index (mod 256) to first char c in text, or anything
 * at all if no occurrences are found. (256 - delta[pos]) is the number
 * of positions to skip forward when match at pos fails.
 * assert(MAX_UCHAR + 1 == 256);
 */
static void pm_preprocess(str, len)
char *str; int len;
{
    char *s = str;
    int i = len;
    while(--i >= 0) {
	int c = (unsigned char)s[i];
	delta[i] = i - alpha[c];
	alpha[c] = i;
    }
}

/*
 * while scanning text from pos. i, we have found a match at mpos.
 * now, print out the processed portion of the text, and modify the
 * alpha[] table. return new pos. to start scanning from.
 */
static int pm_match(str, i, mpos, m)
char *str; int i, mpos; marknode *m;
{
    static char begin[CAPLEN], end[CAPLEN];

    attr_string(m->attrcode, begin, end);
    put_sn(str + i, mpos - i);
    put_s(begin); put_s(m->pattern); put_s(end);

    for(mpos += m->plen; i < mpos; i++)
	alpha[(unsigned char)str[i]] -= delta[i];
    return i;
}

/*
 * print line, scanning for marked text
 *
 * NOTE: perhaps we ought to match a marker starting/ending with
 * a !islabel() char even if it's not preceded/followed by one.
 * NOTE: we may test islabel(str[len]), so this char counts too.
 */
void mark_and_print(str, len)
char *str; int len;
{
    int i;

    if(markers) for(pm_preprocess(str, len), i = 0; ; ) {
	int mpos = -1;
	marknode *m = 0, *np = markers;

	do {
	    register unsigned char *s, *p;
	    int spos, scan = len - np->plen;

	    if(scan < 0) continue;
	    if(scan > (unsigned)mpos) {
		scan = mpos;	/* don't scan past prev. mpos */
		/* only longer pattern is tried at prev. mpos */
		if(np->plen <= m->plen) scan--;
	    }

	    p = (unsigned char *)np->pattern;
	    spos = (unsigned char)(alpha[*p] - i);
	    for(spos += i; spos <= scan; spos += 256) { /* MAX_UCHAR+1 */
		s = (unsigned char *)str + spos;
		if(*s != *p) continue;
		if((!spos || !islabel(s[-1])) && !islabel(s[np->plen])) {
		    do p++, s++; while(*p && *p == *s);
		    if(!*p) {			/* a match! */
			m = np;  mpos = spos;
			break;
		    }
		    p = (unsigned char *)np->pattern;
		}
		spos -= delta[spos];
	    }
	} while((np = np->next));

	if(m) {
	    i = pm_match(str, i, mpos, m);
	    if(i >= len) return;
	} else {
	    len -= i;  str += i;
	    break;
	}
    }
    put_sn(str, len);
}

/*
 * NOTE: We do not perform any complex terminal emulation. But we want to
 * keep track of the last wrapped line (the prompt). We are able to do so
 * if the line is made up of printable chars. In truth, the prompt should
 * not be sustained at all if it contains control chars (with the possible
 * exception of '\b', '\r'). So, we can only hope that (a) the input only
 * contains harmless control sequences, if any, (b) there is no incomplete
 * sequence at the end of the line, and (c) that we don't mark text within
 * control sequence by accident.
 * Bug: incomplete escape sequence at the end of the string is printed...
 * Bug: it is possible that a letter ending an escape sequence is matched
 * as the first char of a marker.
 */

/*
 * write string to stdout, wrapping words to the next line.
 * return pointer to the last wrapped line, or NULL.
 *
 * FEATURE: a feeble attempt is made to handle the csi sequences,
 * so that the attribute changes will probably look ok [in prompt].
 * FEATURE: right now, a marker is not recognized if it gets wrapped.
 * NOTE: perhaps we should (optionally) expand tabs.
 */
char *wrap_out(str)
char *str;
{
    char *tail = 0;
    if(*str) {
	int cc = col0, width = cols_1;
	unsigned char *wp, *s = (unsigned char *)str;
	for(--s, wp = 0; ; ) {
	    if(isgraph(*++s)) {
		if(cc++ < width) continue;
	    } else switch(*s) {
		default:
		    continue;
		case '\033':
		    if(!*++s) break;	/* hmmm */
		    if(*s == '[')
			while(s[1] >= ' ' && *++s < '@');
		    continue;
		case '\b':
		    if(--cc > 0) continue;
		case '\r':
		case '\n':	/* this should also increment line0 */
		    cc = 0;
		    tail = (char *)s + 1;
		case '\013':	/* these should also increment line0 */
		case '\014':
		    wp = 0; continue;
		case '\t':
		    /* tab to screen end might be glitchy, force wrap */
		    if(cc < width && (cc | 7) + 1 < width) {
			cc = (cc | 7) + 1;
			wp = s; continue;
		    }
		    wp = 0; break;
		case ' ':
		    if(cc < width) {
			cc++;
			wp = s; continue;
		    }
		    wp = 0; break;
		case '\0':
		    wp = 0; break;
	    }
	    if(OPTION(OPT_WORDWRAP) && wp) s = wp + 1;
	    mark_and_print(str, (char *)s - str);
	    if(*s == '\t' && cc < width) {
		/* skip if tab crosses end of the line */
		cc = width;
		s++;
	    }
	    if(!*s) break;
	    cc = 0;  newline();
	    tail = str = (char *)s;
	    --s, wp = 0;
	}
	col0 = cc;
    }
    return tail;
}

static char *wrap_pager_line(str)
char *str;
{
    char *s = str;
    int i = cols_1;
    while(*s && *s != '\n' && (++s, --i));
    put_sn(str, s - str);  newline();
    if(OPTION(OPT_PAGERWRAP)) {
	if(*s == '\n') s++;
	if(!*s) s = 0;
    } else s = 0;
    return s;
}

/*
 * pager control. space, enter, 'q' are recognized.
 */
int pager(c)
char c;
{
    int n = 1;
    if(c == '\n' || (c == ' ' && (n = lines_1, 1)))
	pager_lines -= n;
    else if(c == 'q' || (n = 0, 1))
	while(pager_text) delete_pagernode();
    put_s(begoln); put_s(clreoln);
    if(!pager_prompt()) {
	redraw_prompt();
	if(c) redraw_input();
    }
    return n;
}

/*
 * print more pager lines, if there's room. display "--More--" prompt
 */
static int pager_prompt()
{
    defnode *p;
    while((p = pager_text)) {
	if(pager_lines < lines_1) {
	    char *s = p->field;
	    do {
		s = wrap_pager_line(s);
		pager_lines++;
	    } while(s && pager_lines < lines_1);
	    if(!s) {
		delete_pagernode();
		continue;
	    } else {
		char *d = p->field;
		while((*d++ = *s++));
	    }
	}
	printf("%s--More--%s", modeinv, modenorm);
	return 1;
    }
    pager_lines = 0;
    return 0;
}

/*
 * clear the prompt if there is one
 */
void clear_prompt()
{
    if(col0) {
	put_s(begoln);
	put_s(clreoln);
    }
}

/*
 * redisplay the prompt if needed
 */
void redraw_prompt()
{
    if(pager_text && pager_prompt())
	return;
    if(!expect_prompt && col0 && !promptstr[col0]) {
	hilite(0);
#if 0	/* mark_and_print() or wrap_out() ? */
	col0 = 0; wrap_out(promptstr);
#else
	mark_and_print(promptstr, col0);
#endif
    } else
	col0 = 0;
}

/*
 * clear both prompt and input line
 */
void clear_all()
{
    gotoxy(0, line0);
    hilite(0);
    put_s(clreoscr);
}

/*
 * redraw both prompt and input line
 */
void redraw_all()
{
    redraw_prompt();
    redraw_input();
}

/*
 * see whether the window size has changed. return 1 if it has.
 * check_winsize() is always called with sockfd == main_sockfd.
 */
int check_winsize()
{
    /*
     * NOTE: When a window gets smaller, we may need to zero line0
     * (to keep it in window). Subsequently expanding the window
     * can then reveal old text we did not clean up, in case our
     * terminal has memory_above.
     */
    struct winsize z;
    /* if ioctl fails or gives silly values, don't change anything */
    if(ioctl(0, TIOCGWINSZ, &z) == 0 && z.ws_row && z.ws_col > 1 &&
	    (z.ws_row != lines || z.ws_col != cols)) {
	/* NOTE: we assume the bottom line remains on the bottom. */
	line0 += z.ws_row - lines;
	if(line0 < 0) line0 = 0;
	lines = z.ws_row;
	lines_1 = lines - 1;
	cols = z.ws_col;
	cols_1 = cols - !wrapglitch;
	if(col0 >= cols_1) col0 = 0; /* kill the prompt if it's too long */
	send_term_size();
	return 1;
    }
    return 0;
}
